import socket
import collections
import select
import threading
import errno
import uuid
import time
import ssl

from stomp.exceptions import NoReceiptGiven
from stomp.exceptions import StompException
from stomp.datastructures import RawFrame
from stomp.frame import Connect as ConnectFrame
from stomp.frame import Frame
from stomp.const import EOL
from stomp.const import EOF
from stomp.const import ERROR
from stomp.const import RECEIPT
from stomp.const import CONNECTED
from stomp.const import DISCONNECT
from stomp.utils import parse_raw_frames
from stomp.utils import parse_receipt_id


class Transport(object):
    """Handles the sending and receiving of data between
    the STOMP server and the client.

    Handles both incoming and outgoing frames. Starts an
    event loop in the background and writes all data received
    to an internal buffer. The buffer is consumed by calling \
    the :class:`Transport.next() method.`

    Frames are either stored in an internal list or passed
    to a dispatcher (using the :class:`Transport.set_dispatcher()`
    method).
    """
    default_chunk_size = 1024
    chunk_size = property(lambda self: self._chunk_size)
    closed = property(lambda self: self._closed)

    @property
    def messages(self):
        return len(self._frames)
    
    def __init__(self, host, port, encoding='utf-8', chunk_size=1024,
        enable_ssl=False):
        """Initialize a new :class:`Transport` object.
    
        Args:
            host (str): the IP-address or hostname of the STOMP
                server.
            port (int): specifies the port at which the STOMP
                server is listening.
            encoding (str): specifies the encoding used by the 
                connection.
            enable_ssl (bool): enables or disables SSL. Default is
                False.
        """
        self._buffer = ''
        self._lock = threading.Lock()
        self._buffer_lock = threading.Lock()
        self._socket = None
        self._host = host
        self._port = port
        self._encoding = encoding
        self._closed = False
        self._receipts = collections.deque([], 1000)
        self._frames = []
        self._must_exit = False
        self._chunk_size = chunk_size
        self._ssl = None
        if enable_ssl:
            self._ssl = ssl.create_default_context(
            ssl.Purpose.CLIENT_AUTH)
        self._exception = None
        self._dispatcher = None

        # Threading events.
        self.__connected_event = threading.Event()
        self.__disconnected_event = threading.Event()

    def connect(self):
        """Establish a connection to the STOMP server."""
        if self._socket is not None:
            raise EnvironmentError(-1, 
                "Cannot connect in current socket state.")
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self._ssl is not None:
            self._socket = self._ssl.wrap_socket(self._socket)
        self._socket.connect((self._host, self._port))

    def set_dispatcher(self, dispatcher):
        """Sets the dispatcher for received frames."""
        self._dispatcher = dispatcher

    def login(self, vhost, username, password, heartbeats=(0,0)):
        """Login to the remote host.

        Args:
            vhost (str): identifies the virtual host to connect.
            username (str): the username to login with.
            password (str): the password to authenticate with.
            heartbeats (tuple): specifies the heartbeating 
                contract.

        Returns:
            None

        Raises:
            stomp.exceptions.AuthenticationError: invalid credentials.
        """
        if heartbeats is None:
            heartbeats = (0,0)
        frame = ConnectFrame(vhost, username, password,
            server_heartbeats=heartbeats[1],
            client_heartbeats=heartbeats[0])
        self._socket.send(bytes(frame))

        # Now we set the socket to non-blocking to start the main
        # loop of receiving data.
        self._socket.setblocking(0)

        # Assume CONNECT is succesful, and set unset the disconnected
        # event.
        self.__disconnected_event.clear()

    def start(self):
        """Starts the transport.

        Begin with periodically receiving data from the remote STOMP
        server and expose it through the :meth:`Transport.next()`
        method.
        """
        #self._start_timer(0.1, aself._recv)        
        self.__thread = threading.Thread(target=self.__main_event_loop)
        self.__thread.daemon = True
        self.__thread.start()
        self.__ready = False

    def __main_event_loop(self):
        while True:
            # If the transport is closed or some error occurred,
            # we must exit the main event loop.
            if self._must_exit:
                self._closed = True
                return
            try:
                self.__main_event()
            except Exception as e:
                if not hasattr(e, 'errno')\
                or e.errno not in (errno.EAGAIN, errno.ECONNRESET):
                    raise
                if e.errno == errno.ECONNRESET:
                    # Some bad shit happened
                    raise # TODO: handle this.

            # Since the minimal heartbeat interval is 100 milliseconds,
            # I consider is safe to sleep for at least 100 milliseconds,
            # additionally releasing the lock on the socket so data
            # actually be sent to the server.
            time.sleep(0.1)

    def __main_event(self):
        with self._lock:
            buf = self._socket.recv(self.chunk_size)
            
        # Lock the data buffer and parse frames from it,
        # if possible.
        with self._buffer_lock:
            self._write_buffer(buf.decode(encoding=self._encoding))
            frames, self._buffer = parse_raw_frames(self._buffer)
            for frame in map(RawFrame.fromtuple, frames):
                if frame.frame_type == RECEIPT:
                    self._receipts.append(parse_receipt_id(frame.raw_headers))
                    continue
                if frame.frame_type == ERROR:
                    # If the server sent us an ERROR frame, it will close
                    # the connection right after, to we must handle that
                    # gracefully (TODO).
                    error_frame = Frame.fromraw(frame)
                    self._exception = StompException(error_frame,
                        error_frame.headers.get('message'))
                    self._must_exit = True
                if frame.frame_type == CONNECTED:
                    self.__ready = True
                    self.__connected_event.set()
                if self._dispatcher:
                    self._dispatcher(frame)
                else:
                    self._frames.append(frame)

    def wait_for_connection(self, timeout=None):
        """Blocks until a connection with the STOMP server has been
        established.

        Args:
            timeout (int): a timeout, in milliseconds.
        """
        if not self.__connected_event.wait(timeout / 1000):
            raise EnvironmentError(errno.ETIMEDOUT,
                "Connection attempt to STOMP server timed out.")

    def disconnect(self, receipt=True):
        """Sends a ``DISCONNECT`` frame to the STOMP server."""
        self.__disconnected_event.set()
        self.send_frame(Frame(DISCONNECT), receipt=receipt)

    def is_disconnected(self):
        return self.__disconnected_event.is_set()

    def destroy(self):
        """Destroys the :class:`Transport` and cleans up all resources."""
        # If the receiver thread is still running, kill it.
        if self.__thread.is_alive():
            self._must_exit = True
            while self.__thread.is_alive():
                time.sleep(0.1)
        del self._socket
        self.__thread = None
        self._socket = None

    def next(self):
        """Yields the next frame from the local buffer."""
        with self._buffer_lock:
            try:
                raw_frame = self._frames.pop(0)
            except IndexError:
                raise StopIteration
        return raw_frame

    def send_frame(self, frame, receipt=False):
        """Sends a frame to the STOMP server.

        Args:
            frame (stomp.Frame): the frame to send to server.
            receipt (bool): indicates if we want to receive
                a receive from the server. If `receipt` is
                ``True``, meth:`Transport.send` will block
                until a receipt has arrived.
            timeout (int): milliseconds to wait for a receipt.
 
        Raises:
            EnvironmentError: timeout while waiting for receipt.
            StompException: we received an ``ERROR`` frame from the
                server and the connection was closed.
        """
        if self.is_disconnected():
            raise EnvironmentError(errno.ENOTCONN,
                "No session has been established with the STOMP server.")
        if self._exception:
            raise self._exception
        with self._lock:
            receipt_id = uuid.uuid4().hex
            if receipt:
                frame.add_header('receipt', receipt_id)
            try:
                self._socket.setblocking(1)
                self._socket.sendall(bytes(frame))
            finally:
                self._socket.setblocking(0)
        if receipt:
            self._block_until_receipt(receipt_id)
            assert receipt_id not in self._receipts,\
                "Receipts should be removed after verification."

    def _write_buffer(self, buf):
        # Writes a string to the internal buffer. If the length of
        # the received sequence is 1 and EOL, we assume it is a
        # heartbeat.
        if buf == EOL:
            return
        self._buffer += buf

    def __iter__(self):
        while True:
            yield self.next()

    def _block_until_receipt(self, receipt_id, timeout=None,
        poll_interval=100):
        # Block until a receipt with the specified `receipt_id`
        # has been received from the server.
        poll_interval /= 1000
        if timeout is not None:
            timeout /= 1000
        elapsed = 0
        while receipt_id not in self._receipts:
            time.sleep(poll_interval)
            elapsed += poll_interval
            if timeout and (elapsed > timeout):
                raise NoReceiptGiven(
                    "Server did not send us a receipt within the spec"\
                    "ified window of {0}ms".format(timeout * 1000)
                )
        self._receipts.remove(receipt_id)

