import re
import socket
import copy
import collections

from stomp.const import ALLOWED_CHANNEL_PREFIXES
from stomp.const import EOL
from stomp.const import EOF
from stomp.const import SEND
from stomp.const import MESSAGE
from stomp.const import ERROR
from stomp.defaults import DEFAULT_STOMP_PORT


channel_pattern = re.compile(
    '^/({0})/.*$'.format(
        '|'.join(ALLOWED_CHANNEL_PREFIXES)
    )
)
dsn_pattern = re.compile("^tcp://(?P<username>[\d\w]+):(?P<password>[\d\w]+)@(?P<host>[\d\w\.\:]+)/(?P<vhost>.*)$")
content_length_pattern = re.compile('content-length:(\s+)?(?P<content_length>[0-9]+)(\s+)?\n')


def is_valid_channel(channel):
    """Asses if a given channel name is valid. This is hardcore leakage, but
    it will do for now.
    """
    return channel_pattern.match(channel) is not None


ConnectionParams = collections.namedtuple('ConnectionParams',
    ['username','password','host','port','vhost'])


def parse_dsn(dsn):
    """Parses a connection string and returns a dictionary. Missing values,
    such as the port number, will be provided defaults.
    """
    m = dsn_pattern.match(dsn)
    if not m:
        raise ValueError("Invalid Data Source Name: {0}".format(dsn))
    opts = m.groupdict()
    address = opts.pop('host').split(':')
    if len(address) == 1:
        address.append(DEFAULT_STOMP_PORT)
    # try:
    #     address[0] = socket.gethostbyname(address[0])
    # except socket.gaierror as e:
    #     if e.errno != -3:
    #         raise
    #     raise ValueError("Invalid hostname: {0}".format(address[0]))
    address[1] = int(address[1])
    opts['host'], opts['port'] = address
    return ConnectionParams(**opts)


def parse_raw_frames(buf):
    """Parses all frames from stream `buf`."""
    frame, buf = parse_raw_frame_from_stream(buf)
    frames = []
    while frame is not None:
        frame_type, frame = frame.split(EOL, 1)
        parts = frame.split(EOL*2)
        if len(parts) == 1:
            parts.append(EOF)
        head, body = parts
        assert body[-1] == EOF
        frames.append((frame_type, head, body[:-1]))
        frame, buf = parse_raw_frame_from_stream(buf)
    return frames, buf


def parse_raw_frame_from_stream(buf):
    """Parses a complete frame from the stream `buf`, and return a tuple
    holding the frame and the remainder. If no frame could be parsed,
    the first member is None.
    """
    ptr = buf.find(EOL)
    if ptr == -1 or (EOF not in buf):
        return (None, buf)
    frame_type = buf[:ptr].strip(EOL)

    # From the spec:
    #
    # If a frame body is present, the SEND, MESSAGE and ERROR frames 
    # SHOULD include a content-length header to ease frame parsing. 
    # If the frame body contains NULL octets, the frame MUST include 
    # a content-length header.
    if frame_type not in (MESSAGE, SEND, ERROR):   
        raw_frame = buf[:buf.find(EOF) + 1]
    else:
        # We must find the content-length header and parse the
        # length of the message.
        header = buf[:buf.find(EOL*2) + 2]
        sre = content_length_pattern.search(header, re.DOTALL)
        if sre is not None:
            content_length = int(sre.group('content_length'))
            raw_frame = buf[: len(header) + content_length + 1]

            # Ensure that the message length is smaller than or equal
            # to the buffer, that a complete frame is parsed.
            message_length = len(header) + content_length + 1
            if len(buf) < message_length:
                # Ugly early return.
                return None, buf
        else:
            # This might go very wrong if the server sends us
            # malformed messages (TODO).
            raw_frame = buf[:buf.find(EOF) + 1]
    buf = buf[len(raw_frame):]

    assert raw_frame[-1] == EOF, "Correctly parsed messages end with EOL"

    # Remove any newlines from the beginning of the buffer. Shouldn't
    # be there, but reality showed that they are. TODO: find out why.
    while buf and (buf[0] == EOL):
        buf = buf[1:]

    return (raw_frame.strip(EOL), buf)


def parse_receipt_id(raw_header):
    """Parses the receipt-id from a raw header.

    Args:
        raw_frame (str): a character sequences holding the 
            headers of a STOMP frame.

    Returns:
        str

    Raises:
        ValueError: no receipt-id header present.
    """
    ptr = raw_header.index('receipt-id:')
    if ptr == -1:
        raise ValueError("Malformed frame headers: {0}".format(raw_header))
    header = raw_header[ptr:][raw_header[ptr:].find(EOL)]
    return raw_header[raw_header.index(':') + 1:]
